@@ -26,7 +26,7 @@ fi
 
 set -e
 
-BUILD_DIR="build.tmp"
+BUILD_DIR="build-kf5.tmp"
 rm -fR "$BUILD_DIR"
 mkdir -p "$BUILD_DIR"
 
@@ -36,7 +36,7 @@ pushd "$(realpath $BUILD_DIR)"
 echo "INFO: Installation prefix: $PREFIX"
 echo "INFO: Build type         : $BUILD_TYPE"
 
-cmake -DKS_KF5=true -DCMAKE_BUILD_TYPE=$BUILD_TYPE -DCMAKE_INSTALL_PREFIX="$PREFIX" ..
+cmake -DKS_FORCE_KF5=true -DKS_KF5=true -DCMAKE_BUILD_TYPE=$BUILD_TYPE -DCMAKE_INSTALL_PREFIX="$PREFIX" ..
 make -j2
 
 popd
