@@ -1,19 +1,5 @@
-// kshutdown.cpp - KShutdown base library
-// Copyright (C) 2007  Konrad Twardowski
-//
-// This program is free software; you can redistribute it and/or modify
-// it under the terms of the GNU General Public License as published by
-// the Free Software Foundation; either version 2 of the License, or
-// (at your option) any later version.
-//
-// This program is distributed in the hope that it will be useful,
-// but WITHOUT ANY WARRANTY; without even the implied warranty of
-// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-// GNU General Public License for more details.
-//
-// You should have received a copy of the GNU General Public License along
-// with this program; if not, write to the Free Software Foundation, Inc.,
-// 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
+// SPDX-License-Identifier: GPL-2.0-or-later
+// SPDX-FileCopyrightText: 2007 Konrad Twardowski
 
 #include "kshutdown.h"
 
@@ -28,6 +14,7 @@
 #include "actions/lock.h"
 
 #include <QDebug>
+#include <QFormLayout>
 #include <QLineEdit>
 #include <QMessageBox>
 #include <QPushButton>
@@ -39,373 +26,22 @@
 	#include <windows.h>
 	#include <powrprof.h>
 #else
-	#include <csignal> // for ::kill
-
 	#include <QThread>
 #endif // Q_OS_WIN32
 
-using namespace KShutdown;
-
-bool Action::m_totalExit = false;
-#ifdef QT_DBUS_LIB
-QDBusInterface *Action::m_loginInterface = nullptr;
-
-QDBusInterface *PowerAction::m_halDeviceInterface = nullptr;
-QDBusInterface *PowerAction::m_halDeviceSystemPMInterface = nullptr;
-QDBusInterface *PowerAction::m_upowerInterface = nullptr;
-
-bool StandardAction::m_kdeShutDownAvailable = false;
-QDBusInterface *StandardAction::m_consoleKitInterface = nullptr;
-QDBusInterface *StandardAction::m_kdeSessionInterface = nullptr;
-QDBusInterface *StandardAction::m_lxqtSessionInterface = nullptr;
-QDBusInterface *StandardAction::m_razorSessionInterface = nullptr;
-#endif // QT_DBUS_LIB
-
-// DateTimeEdit
-
-// public:
-
-DateTimeEdit::DateTimeEdit()
-	: QDateTimeEdit()
-{
-	connect(
-		lineEdit(), SIGNAL(selectionChanged()),
-		SLOT(onLineEditSelectionChange())
-	);
-}
-
-DateTimeEdit::~DateTimeEdit() = default;
-
-// private slots:
-
-void DateTimeEdit::onLineEditSelectionChange() {
-	if (displayedSections() != (HourSection | MinuteSection))
-		return;
-
-	QString selectedText = lineEdit()->selectedText();
-	// HACK: Change selection from "15h" to "15" after double click.
-	//       This fixes Up/Down keys and mouse wheel editing.
-	if (selectedText == time().toString("hh'h'"))
-		lineEdit()->setSelection(0, 2);
-	else if (selectedText == time().toString("mm'm'"))
-		lineEdit()->setSelection(6, 2);
-}
-
-// DateTimeTriggerBase
-
-// public
-
-DateTimeTriggerBase::DateTimeTriggerBase(const QString &text, const QString &iconName, const QString &id) :
-	Trigger(text, iconName, id),
-	m_dateTime(QDateTime()),
-	m_endDateTime(QDateTime())
-{
-}
-
-DateTimeTriggerBase::~DateTimeTriggerBase() {
-	delete m_edit;
-	m_edit = nullptr;
-}
-
-bool DateTimeTriggerBase::canActivateAction() {
-	QDateTime now = QDateTime::currentDateTime();
-	int secsTo;
-	m_status = createStatus(now, secsTo);
-
-	if (secsTo > 0) {
-		MainWindow *mainWindow = MainWindow::self();
-		mainWindow->progressBar()->setValue(secsTo);
-		
-		QString id = QString::null;
-		if ((secsTo < 60) && (secsTo > 55))
-			id = "1m";
-		else if ((secsTo < 300) && (secsTo > 295))
-			id = "5m";
-		else if ((secsTo < 1800) && (secsTo > 1795))
-			id = "30m";
-		else if ((secsTo < 3600) && (secsTo > 3595))
-			id = "1h";
-		else if ((secsTo < 7200) && (secsTo > 7195))
-			id = "2h";
-		if (!id.isNull()) {
-			emit notify(id, mainWindow->getDisplayStatus(MainWindow::DISPLAY_STATUS_HTML));
-		}
-	}
-	
-	return now >= m_endDateTime;
-}
-
-QWidget *DateTimeTriggerBase::getWidget() {
-	if (!m_edit) {
-		m_edit = new DateTimeEdit();
-		connect(m_edit, SIGNAL(dateChanged(const QDate &)), SLOT(syncDateTime()));
-		connect(m_edit, SIGNAL(timeChanged(const QTime &)), SLOT(syncDateTime()));
-		m_edit->setObjectName("date-time-edit");
-
-		// larger font
-		Utils::setFont(m_edit, 2, true);
-	}
-
-	return m_edit;
-}
-
-QString DateTimeTriggerBase::longDateTimeFormat() {
-	QString dateFormat = "MMM d dddd";
-	QString timeFormat = QLocale::system()
-		.timeFormat(QLocale::ShortFormat);
-
-	if (timeFormat == "h:mm:ss AP")
-		timeFormat = "h:mm AP";
-	else
-		timeFormat = "hh:mm";
-	
-	return dateFormat + ' ' + timeFormat;
-}
-
-void DateTimeTriggerBase::readConfig(Config *config) {
-	m_dateTime = config->read("Date Time", m_dateTime).toDateTime();
-}
-
-void DateTimeTriggerBase::writeConfig(Config *config) {
-	if (m_edit)
-		config->write("Date Time", m_edit->dateTime());
-}
-
-void DateTimeTriggerBase::setDateTime(const QDateTime &dateTime) {
-	m_dateTime = dateTime;
-	m_edit->setDateTime(dateTime);
-}
-
-void DateTimeTriggerBase::setState(const State state) {
-	if (state == State::Start) {
-		m_endDateTime = calcEndTime();
-		
-		// reset progress bar
-		if (m_supportsProgressBar) {
-			QDateTime now = QDateTime::currentDateTime();
-			int secsTo = now.secsTo(m_endDateTime);
-			ProgressBar *progressBar = MainWindow::self()->progressBar();
-			progressBar->setTotal(secsTo);
-			progressBar->setValue(-1);
-		}
-	}
-}
-
-// protected
-
-void DateTimeTriggerBase::updateStatus() {
-	QDateTime now = QDateTime::currentDateTime();
-	int secsTo;
-
-	m_dateTime = m_edit->dateTime();
-	m_endDateTime = calcEndTime();
-	m_status = createStatus(now, secsTo);
-}
-
-// private slots
-
-void DateTimeTriggerBase::syncDateTime() {
-	updateStatus();
-	emit statusChanged(false);
-}
-
-// private
-
-QString DateTimeTriggerBase::createStatus(const QDateTime &now, int &secsTo) {
-	m_statusType = InfoWidget::Type::Info;
-	
-	secsTo = now.secsTo(m_endDateTime);
-	if (secsTo > 0) {
-		const int DAY = 86'400;
-		
-		QString result;
-		if (secsTo < DAY) {
-			result = '+' + QTime(0, 0).addSecs(secsTo).toString(LONG_TIME_DISPLAY_FORMAT);
-		}
-		else {
-			result += "24h+";
-		}
-		
-		result += " (";
-
-		if (secsTo >= DAY)
-			result += i18n("not recommended") + ", ";
-
-// TODO: do not bold effective time
-		result += i18n("selected time: %0").arg(m_endDateTime.toString(longDateTimeFormat()));
-		result += ')';
-		
-		return result;
-	}
-	else if (secsTo == 0) {
-		return QString::null;
-	}
-	else /* if (secsTo < 0) */ {
-		m_statusType = InfoWidget::Type::Warning;
-	
-		return i18n("Invalid date/time");
-	}
-}
-
-// DateTimeTrigger
-
-// public
-
-DateTimeTrigger::DateTimeTrigger() :
-	DateTimeTriggerBase(i18n("At Date/Time"), "view-pim-calendar", "date-time")
-{
-	setCanBookmark(true);
-
-	m_dateTime = QDateTime::currentDateTime().addSecs(60 * 60/* hour */); // set default
-	m_supportsProgressBar = true;
-}
-
-QDateTime DateTimeTrigger::dateTime() {
-	DateTimeTriggerBase::getWidget();
-
-	return m_edit->dateTime();
-}
-
-QString DateTimeTrigger::getStringOption() {
-	if (!m_edit)
-		return QString::null;
-	
-	return m_edit->time().toString(TIME_PARSE_FORMAT);
-}
-
-void DateTimeTrigger::setStringOption(const QString &option) {
-	if (!m_edit)
-		return;
-
-	QDate date;
-	QTime time = QTime::fromString(option, TIME_PARSE_FORMAT);
-
-	// select next day if time is less than current time
-	if (time < QTime::currentTime())
-		date = QDate::currentDate().addDays(1);
-	else
-		date = m_edit->date();
-	m_edit->setDateTime(QDateTime(date, time));
-}
-
-QWidget *DateTimeTrigger::getWidget() {
-	bool initDateTime = !m_edit;
-
-	DateTimeTriggerBase::getWidget();
-
-	m_edit->setCalendarPopup(true);
-	
-	// Fix for BUG #2444169 - remember the previous shutdown settings
-	if (initDateTime) {
-		QDate date = m_dateTime.date();
-		QTime time = m_dateTime.time();
-
-		// select next day if time is less than current time
-		if ((date == QDate::currentDate()) && (time < QTime::currentTime()))
-			date = date.addDays(1);
-		m_edit->setDateTime(QDateTime(date, time));
-	}
-	
-	m_edit->setDisplayFormat(DateTimeTriggerBase::longDateTimeFormat());
-	m_edit->setMinimumDate(QDate::currentDate());
-	//m_edit->setMinimumDateTime(QDateTime::currentDateTime());
-	m_edit->setToolTip(i18n("Enter date and time"));
-
-	return m_edit;
-}
-
-void DateTimeTrigger::setState(const State state) {
-	DateTimeTriggerBase::setState(state);
-	
-	if (state == State::InvalidStatus) {
-		// show warning if selected date/time is invalid
-		if (QDateTime::currentDateTime() >= dateTime()) {
-			m_status = i18n("Invalid date/time");
-			m_statusType = InfoWidget::Type::Warning;
-		}
-		else {
-			updateStatus();
-		}
-	}
-}
-
-// protected
-
-QDateTime DateTimeTrigger::calcEndTime() {
-	return m_edit->dateTime();
-}
-
-// NoDelayTrigger
-
-// public
-
-NoDelayTrigger::NoDelayTrigger() :
-	Trigger(i18n("No Delay"), "dialog-warning", "no-delay") {
-	
-	setCanBookmark(true);
-}
-
-// TimeFromNowTrigger
-
-// public
-
-TimeFromNowTrigger::TimeFromNowTrigger() :
-	DateTimeTriggerBase(i18n("Time From Now (HH:MM)"), "chronometer", "time-from-now")
-{
-	setCanBookmark(true);
-
-	m_dateTime.setTime(QTime(1, 0, 0)); // set default
-	m_supportsProgressBar = true;
-}
-
-QString TimeFromNowTrigger::getStringOption() {
-	if (!m_edit)
-		return QString::null;
-
-	return m_edit->time().toString(TIME_PARSE_FORMAT);
-}
-
-void TimeFromNowTrigger::setStringOption(const QString &option) {
-	if (!m_edit)
-		return;
-
-	QTime time = QTime::fromString(option, TIME_PARSE_FORMAT);
-	m_edit->setTime(time);
-}
-
-QWidget *TimeFromNowTrigger::getWidget() {
-	DateTimeTriggerBase::getWidget();
-
-	m_edit->setDisplayFormat(TIME_DISPLAY_FORMAT);
-	m_edit->setTime(m_dateTime.time());
-	m_edit->setToolTip(i18n("Enter delay in \"HH:MM\" format (Hour:Minute)"));
-
-	return m_edit;
-}
-
-// protected
-
-QDateTime TimeFromNowTrigger::calcEndTime() {
-	QTime time = m_dateTime.time();
-	int m = (time.hour() * 60) + time.minute();
-	
-	return QDateTime::currentDateTime().addSecs(m * 60);
-}
-
 // PowerAction
 
 // public
 
 PowerAction::PowerAction(const QString &text, const QString &iconName, const QString &id) :
-	Action(text, iconName, id),
-	m_methodName(QString::null) {
-	
+	Action(text, iconName, id)
+{
 	setCanBookmark(true);
 }
 
 #ifdef QT_DBUS_LIB
 QDBusInterface *PowerAction::getHalDeviceInterface() {
-	if (!m_halDeviceInterface) {
+	if (m_halDeviceInterface == nullptr) {
 		// DOC: http://people.freedesktop.org/~dkukawka/hal-spec-git/hal-spec.html
 		m_halDeviceInterface = new QDBusInterface(
 			"org.freedesktop.Hal",
@@ -423,7 +59,7 @@ QDBusInterface *PowerAction::getHalDeviceInterface() {
 }
 
 QDBusInterface *PowerAction::getHalDeviceSystemPMInterface() {
-	if (!m_halDeviceSystemPMInterface) {
+	if (m_halDeviceSystemPMInterface == nullptr) {
 		// DOC: http://people.freedesktop.org/~dkukawka/hal-spec-git/hal-spec.html
 		m_halDeviceSystemPMInterface = new QDBusInterface(
 			"org.freedesktop.Hal",
@@ -440,7 +76,7 @@ QDBusInterface *PowerAction::getHalDeviceSystemPMInterface() {
 
 QDBusInterface *PowerAction::getUPowerInterface() {
 	// DOC: http://upower.freedesktop.org/docs/UPower.html
-	if (!m_upowerInterface) {
+	if (m_upowerInterface == nullptr) {
 		m_upowerInterface = new QDBusInterface(
 			"org.freedesktop.UPower",
 			"/org/freedesktop/UPower",
@@ -486,8 +122,8 @@ bool PowerAction::onAction() {
 	return false;
 #else
 	// lock screen before hibernate/suspend
-	if (Config::lockScreenBeforeHibernateVar) {
-		LockAction::self()->activate(false);
+	if (Config::lockScreenBeforeHibernate) {
+		LockAction::self()->activate();
 
 		// HACK: wait for screensaver
 		QThread::sleep(1);
@@ -549,7 +185,7 @@ bool PowerAction::onAction() {
 
 // protected
 
-bool PowerAction::isAvailable(const PowerActionType feature) const {
+bool PowerAction::isAvailable(const PowerActionType feature) {
 #ifdef Q_OS_WIN32
 	if (feature == PowerActionType::Hibernate)
 		return ::IsPwrHibernateAllowed();
@@ -571,13 +207,16 @@ bool PowerAction::isAvailable(const PowerActionType feature) const {
 				QDBusReply<QString> reply = login->call("CanSuspend");
 				qDebug() << "systemd: CanSuspend: " << reply;
 				
-				return reply.isValid() && (reply.value() == "yes");
+				// HACK: Sometimes "challenge" is returned shortly after a Desktop Environment start.
+				// DOC: https://www.freedesktop.org/wiki/Software/systemd/logind/
+// TODO: update actions after "CanXX" change
+				return reply.isValid() && ((reply.value() == "yes") || (reply.value() == "challenge"));
 			} break;
 			case PowerActionType::Hibernate: {
 				QDBusReply<QString> reply = login->call("CanHibernate");
 				qDebug() << "systemd: CanHibernate: " << reply;
 				
-				return reply.isValid() && (reply.value() == "yes");
+				return reply.isValid() && ((reply.value() == "yes") || (reply.value() == "challenge"));
 			} break;
 		}
 	}
@@ -635,16 +274,16 @@ bool PowerAction::isAvailable(const PowerActionType feature) const {
 
 HibernateAction::HibernateAction() :
 	PowerAction(
-		getDisplayName(ActionType::HIBERNATE),
+		i18n("Hibernate"),
 		"system-suspend-hibernate", "hibernate"
 ) {
 	m_methodName = "Hibernate";
 	if (!isAvailable(PowerActionType::Hibernate))
-		disable(i18n("Cannot hibernate computer"));
+		setEnabled(false, i18n("Cannot hibernate computer"));
 
 	setCommandLineOption({ "H", "hibernate" });
 
-	uiAction()->setStatusTip(
+	uiAction()->setToolTip(
 		i18n("Hibernate Computer") + "\n" +
 		i18n("Save the contents of RAM to disk\nthen turn off the computer.")
 	);
@@ -656,33 +295,48 @@ HibernateAction::HibernateAction() :
 
 SuspendAction::SuspendAction() :
 	PowerAction(
-		getDisplayName(ActionType::SUSPEND),
+		i18n("Sleep"), // NOTE: Matches https://phabricator.kde.org/D19184
 		"system-suspend", "suspend") {
 	m_methodName = "Suspend";
 	if (!isAvailable(PowerActionType::Suspend))
-		disable(i18n("Cannot suspend computer"));
+		setEnabled(false, i18n("Cannot suspend computer"));
 
 	setCommandLineOption({ "S", "sleep", "suspend" });
 
-	uiAction()->setStatusTip(
+	uiAction()->setToolTip(
 		i18n("Suspend Computer") + "\n" +
 		i18n("Enter in a low-power state mode.")
 	);
 }
 
-// StandardAction
+// ExitAction
 
 // public
 
-StandardAction::StandardAction(const QString &text, const QString &iconName, const QString &id, const UShutdownType type) :
+ExitAction::ExitAction(const QString &text, const QString &iconName, const QString &id, const UShutdownType type) :
 	Action(text, iconName, id),
 	m_type(type) {
 	
 	setCanBookmark(true);
 
+	m_forceCheckBox = new QCheckBox(i18n("Do not save session / Force shutdown"));
+	m_forceCheckBox->setObjectName("force");
+
+	connect(m_forceCheckBox, &QCheckBox::clicked, [this]() {
+		if (m_forceCheckBox->isChecked()) {
+			UMessageBuilder message(UMessageBuilder::Type::Question);
+			message.okText(i18n("Enable"));
+			message.showKey("KShutdown Force Exit");
+			message.plainText(i18n("Are you sure you want to enable this option?\n\nData in all unsaved documents will be lost!"));
+
+			if (! message.exec(m_forceCheckBox))
+				m_forceCheckBox->setChecked(false);
+		}
+	});
+
 // TODO: clean up kshutdown.cpp, move this to LogoutAction
 	#ifdef QT_DBUS_LIB
-	if (!m_kdeSessionInterface) {
+	if (m_kdeSessionInterface == nullptr) {
 		m_kdeSessionInterface = new QDBusInterface(
 			"org.kde.ksmserver",
 			"/KSMServer",
@@ -692,55 +346,33 @@ StandardAction::StandardAction(const QString &text, const QString &iconName, con
 		m_kdeShutDownAvailable = reply.isValid() && reply.value();
 
 		if (Utils::isKDEFullSession() && !m_kdeShutDownAvailable && (type != U_SHUTDOWN_TYPE_LOGOUT))
-			qDebug() << "No KDE shutdown API available (check \"Offer shutdown options\" in the \"Session Management\" settings)";
+			qDebug() << "No org.kde.ksmserver API available";
 	}
 
-	if (Utils::isRazor() && (type == U_SHUTDOWN_TYPE_LOGOUT)) {
-		m_razorSessionInterface = new QDBusInterface(
-			"org.razorqt.session",
-			"/RazorSession",
-			"org.razorqt.session"
+	if (Utils::kde && (m_kdeShutdownInterface == nullptr)) {
+		m_kdeShutdownInterface = new QDBusInterface(
+			"org.kde.Shutdown",
+			"/Shutdown", // FUN FACT: missing "," causes app crash
+			"org.kde.Shutdown"
 		);
-		QDBusReply<bool> reply = m_razorSessionInterface->call("canLogout");
-		if (!reply.isValid() || !reply.value()) {
-			delete m_razorSessionInterface;
-			m_razorSessionInterface = nullptr;
-			disable("No Razor-qt session found");
-		}
-	}
-
-	if (Utils::isLXQt() && (type == U_SHUTDOWN_TYPE_LOGOUT)) {
-		m_lxqtSessionInterface = new QDBusInterface(
-			"org.lxqt.session",
-			"/LXQtSession",
-			"org.lxqt.session"
-		);
-		QDBusReply<bool> reply = m_lxqtSessionInterface->call("canLogout");
-		if (!reply.isValid() || !reply.value()) {
-			delete m_lxqtSessionInterface;
-			m_lxqtSessionInterface = nullptr;
-			disable("No LXQt session found");
-		}
+		if (m_kdeShutdownInterface->isValid())
+			qDebug() << "Using org.kde.Shutdown API";
+		else
+			qWarning() << "No org.kde.Shutdown API available";
 	}
 	#endif // QT_DBUS_LIB
-
-	#ifndef Q_OS_WIN32
-	m_lxsession = 0;
-	if (Utils::isLXDE() && (type == U_SHUTDOWN_TYPE_LOGOUT)) {
-		bool ok = false;
-		int i = qgetenv("_LXSESSION_PID").toInt(&ok);
-		if (ok) {
-			m_lxsession = i;
-			qDebug() << "LXDE session found: " << m_lxsession;
-		}
-		else {
-			disable("No lxsession found");
-		}
-	}
-	#endif // !Q_OS_WIN32
 }
 
-bool StandardAction::onAction() {
+void ExitAction::initContainerWidget() {
+	makeFormLayout();
+
+	#ifdef Q_OS_WIN32
+	auto *layout = getContainerForm();
+	layout->addRow(m_forceCheckBox);
+	#endif // Q_OS_WIN32
+}
+
+bool ExitAction::onAction() {
 	m_totalExit = true;
 #ifdef Q_OS_WIN32
 	UINT flags = 0;
@@ -814,9 +446,8 @@ bool StandardAction::onAction() {
 		(m_type == U_SHUTDOWN_TYPE_HALT) ||
 		(m_type == U_SHUTDOWN_TYPE_REBOOT)
 	) {
-		BOOL bForceAppsClosed = m_force ? TRUE : FALSE;
-		m_force = false;
-		
+		BOOL bForceAppsClosed = m_forceCheckBox->isChecked() ? TRUE : FALSE;
+
 		BOOL bRebootAfterShutdown = (m_type == U_SHUTDOWN_TYPE_REBOOT) ? TRUE : FALSE;
 		
 		if (::InitiateSystemShutdown(
@@ -848,9 +479,8 @@ bool StandardAction::onAction() {
 	}
 	else {
 		flags += EWX_FORCEIFHUNG;
-		if (m_force) {
+		if (m_forceCheckBox->isChecked()) {
 			flags += EWX_FORCE;
-			m_force = false;
 		}
 
 		if (::ExitWindowsEx(flags, SHTDN_REASON_MAJOR_APPLICATION | SHTDN_REASON_FLAG_PLANNED) == 0) {
@@ -868,7 +498,7 @@ bool StandardAction::onAction() {
 		QStringList args;
 		args << "-r";
 		
-		if (launch("shutdown", args)) {
+		if (Utils::run("shutdown", args)) {
 			// HACK: fixes "Asking kshutdown to quit" message
 			QApplication::quit();
 
@@ -878,7 +508,7 @@ bool StandardAction::onAction() {
 	else if (m_type == U_SHUTDOWN_TYPE_HALT) {
 		QStringList args;
 		
-		if (launch("shutdown", args)) {
+		if (Utils::run("shutdown", args)) {
 			// HACK: fixes "Asking kshutdown to quit" message
 			QApplication::quit();
 
@@ -890,91 +520,23 @@ bool StandardAction::onAction() {
 	
 #else
 
-	// Cinnamon
+	if (strategyManager()->run())
+		return true;
 
-	if (Utils::isCinnamon()) {
-		if (m_type == U_SHUTDOWN_TYPE_LOGOUT) {
-			QStringList args;
-			args << "--logout";
-			args << "--no-prompt";
-
-			if (launch("cinnamon-session-quit", args, true))
-				return true;
-		}
-	}
-	
 	// GNOME Shell, Unity
 
-	else if (Utils::isGNOME() || Utils::isUnity()) {
+	if (Utils::isGNOME() || Utils::isUnity()) {
 		if (m_type == U_SHUTDOWN_TYPE_LOGOUT) {
 			QStringList args;
 // TODO: --force
 			args << "--logout";
 			args << "--no-prompt";
 
-			if (launch("gnome-session-quit", args, true))
+			if (Utils::run("gnome-session-quit", args))
 				return true;
 		}
 	}
 
-	// LXDE
-
-	else if (Utils::isLXDE()) {
-		if (m_type == U_SHUTDOWN_TYPE_LOGOUT) {
-			#ifndef Q_OS_WIN32
-			if (m_lxsession && (::kill(m_lxsession, SIGTERM) == 0))
-				return true;
-			#else
-				return false;
-			#endif // !Q_OS_WIN32
-		}
-	}
-
-	// LXQt
-
-	#ifdef QT_DBUS_LIB
-	else if (Utils::isLXQt()) {
-		if ((m_type == U_SHUTDOWN_TYPE_LOGOUT) && m_lxqtSessionInterface && m_lxqtSessionInterface->isValid()) {
-			QDBusReply<void> reply = m_lxqtSessionInterface->call("logout");
-
-			if (reply.isValid())
-				return true;
-		}
-	}
-/* DEAD: interactive
-			QStringList args;
-			args << "--logout";
-
-			if (launch("lxqt-leave", args))
-				return true;
-*/
-	#endif // QT_DBUS_LIB
-
-	// MATE
-
-	else if (Utils::isMATE()) {
-		if (m_type == U_SHUTDOWN_TYPE_LOGOUT) {
-			QStringList args;
-			args << "--logout";
-
-			if (launch("mate-session-save", args, true))
-				return true;
-		}
-	}
-
-	// Razor-qt
-	
-	#ifdef QT_DBUS_LIB
-	else if (Utils::isRazor()) {
-		if ((m_type == U_SHUTDOWN_TYPE_LOGOUT) && m_razorSessionInterface && m_razorSessionInterface->isValid()) {
-			QDBusReply<void> reply = m_razorSessionInterface->call("logout");
-
-			if (reply.isValid())
-				return true;
-		}
-	}
-	#endif // QT_DBUS_LIB
-
 	// Trinity
 
 	else if (Utils::isTrinity()) {
@@ -989,33 +551,29 @@ bool StandardAction::onAction() {
 			args << "0"; // logout
 			args << "2"; // force now
 
-			if (launch("dcop", args))
+			if (Utils::run("dcop", args))
 				return true;
 		}
 	}
 
 	// Xfce
 
-	else if (Utils::isXfce()) {
+/* TODO: ?
+	else if (Utils::xfce) {
 		switch (m_type) {
 			case U_SHUTDOWN_TYPE_LOGOUT: {
-				QStringList args;
-				args << "--logout";
-
-				if (launch("xfce4-session-logout", args, true))
-					return true;
 			} break;
-/*
+
 			case U_SHUTDOWN_TYPE_REBOOT: {
 				QStringList args;
 				args << "--reboot";
-				if (launch("xfce4-session-logout", args))
+				if (Utils::run("xfce4-session-logout", args))
 					return true;
 			} break;
 			case U_SHUTDOWN_TYPE_HALT: {
 				QStringList args;
 				args << "--halt";
-				if (launch("xfce4-session-logout", args))
+				if (Utils::run("xfce4-session-logout", args))
 					return true;
 			} break;
 			default:
@@ -1023,20 +581,6 @@ bool StandardAction::onAction() {
 
 				return false; // do nothing
 */
-		}
-	}
-
-	// Enlightenment
-
-	else if (Utils::isEnlightenment()) {
-// TODO: use D-Bus
-		if (m_type == U_SHUTDOWN_TYPE_LOGOUT) {
-			QStringList args;
-			args << "-exit";
-			if (launch("enlightenment_remote", args))
-				return true;
-		}
-	}
 
 	// native KDE shutdown API
 
@@ -1051,7 +595,6 @@ bool StandardAction::onAction() {
 			m_type,
 			2 // KWorkSpace::ShutdownModeForceNow
 		);
-
 /* TODO: error checking?
 		if (reply.isValid())
 			return true;
@@ -1072,7 +615,7 @@ bool StandardAction::onAction() {
 			QStringList args;
 			args << "--exit";
 
-			if (launch("openbox", args))
+			if (Utils::run("openbox", args))
 				return true;
 		}
 	}
@@ -1081,9 +624,6 @@ bool StandardAction::onAction() {
 	
 	#ifdef QT_DBUS_LIB
 
-// TODO: move
-	MainWindow::self()->writeConfig();
-	
 	// try systemd/logind
 
 	QDBusInterface *login = getLoginInterface();
@@ -1103,13 +643,13 @@ bool StandardAction::onAction() {
 
 	// try ConsoleKit
 	
-	if ((m_type == U_SHUTDOWN_TYPE_HALT) && m_consoleKitInterface && m_consoleKitInterface->isValid()) {
+	if ((m_type == U_SHUTDOWN_TYPE_HALT) && (m_consoleKitInterface != nullptr) && m_consoleKitInterface->isValid()) {
 		QDBusReply<void> reply = m_consoleKitInterface->call("Stop");
 
 		if (reply.isValid())
 			return true;
 	}
-	else if ((m_type == U_SHUTDOWN_TYPE_REBOOT) && m_consoleKitInterface && m_consoleKitInterface->isValid()) {
+	else if ((m_type == U_SHUTDOWN_TYPE_REBOOT) && (m_consoleKitInterface != nullptr) && m_consoleKitInterface->isValid()) {
 		QDBusReply<void> reply = m_consoleKitInterface->call("Restart");
 
 		if (reply.isValid())
@@ -1143,7 +683,7 @@ bool StandardAction::onAction() {
 // protected
 
 #ifdef QT_DBUS_LIB
-void StandardAction::checkAvailable(const QString &consoleKitName) {
+void ExitAction::checkAvailable(const QString &consoleKitName) {
 	bool available = false;
 	QString error = "";
 // TODO: clean up; return bool
@@ -1160,7 +700,7 @@ void StandardAction::checkAvailable(const QString &consoleKitName) {
 	// try ConsoleKit
 
 	if (!consoleKitName.isEmpty()) {
-		if (!m_consoleKitInterface) {
+		if (m_consoleKitInterface == nullptr) {
 			m_consoleKitInterface = new QDBusInterface(
 				"org.freedesktop.ConsoleKit",
 				"/org/freedesktop/ConsoleKit/Manager",
@@ -1206,7 +746,7 @@ void StandardAction::checkAvailable(const QString &consoleKitName) {
 	
 	// BUG #19 - disable only if both ConsoleKit and native KDE API is unavailable
 	if (!available && !m_kdeShutDownAvailable)
-		disable(error);
+		setEnabled(false, error);
 }
 #endif // QT_DBUS_LIB
 
@@ -1215,17 +755,78 @@ void StandardAction::checkAvailable(const QString &consoleKitName) {
 // public
 
 LogoutAction::LogoutAction() :
-	StandardAction(
-		getDisplayName(ActionType::LOGOUT),
+	ExitAction(
+		#ifdef Q_OS_WIN32
+		i18n("Log Off"),
+		#else
+		i18n("Logout"),
+		#endif // Q_OS_WIN32
 		"system-log-out", "logout", U_SHUTDOWN_TYPE_LOGOUT
 ) {
-	uiAction()->setStatusTip(i18n("End Session"));
+	QString currentUser = Utils::getUser();
+	
+	if (currentUser.isEmpty())
+		currentUser = "?";
 
-	setCommandLineOption({ "l", "logoff", "logout" }, originalText() + " — " + i18n("End Session"));
+	uiAction()->setToolTip(
+		i18n("End Session") + "\n" +
+		"\n" +
+		i18n("Current User: %0").arg(currentUser)
+	);
+
+	setCommandLineOption({ "l", "logoff", "logout" }, Utils::makeTitle(originalText(), i18n("End Session")));
 
 	#ifdef Q_OS_HAIKU
-	disable("");
+	setEnabled(false, "");
 	#endif // Q_OS_HAIKU
+
+	auto sm = strategyManager();
+
+	// by desktop environment
+
+	#ifdef QT_DBUS_LIB
+	if (m_kdeShutdownInterface != nullptr)
+		sm->sessionBus(Utils::kde, m_kdeShutdownInterface, "logout");
+	#endif // QT_DBUS_LIB
+
+	QString lxdeDummy = "Terminate $_LXSESSION_PID PID";
+	if (Utils::lxde) {
+		if (auto pid = Utils::toInt64(qEnvironmentVariable("_LXSESSION_PID"))) {
+			sm->terminate(Utils::lxde, pid.value());
+		}
+		else {
+			sm->dummy(Utils::lxde, lxdeDummy);
+			setEnabled(false, "No LXDE session found");
+		}
+	}
+	else {
+		sm->dummy(Utils::lxde, lxdeDummy);
+	}
+
+	sm->program(Utils::cinnamon, "cinnamon-session-quit", { "--logout", "--no-prompt" });
+
+// TODO: use D-Bus
+	sm->program(Utils::enlightenment, "enlightenment_remote", { "-exit" });
+
+	sm->program(Utils::mate, "mate-session-save", { "--logout" });
+	sm->program(Utils::xfce, "xfce4-session-logout", { "--logout" });
+
+	#ifdef QT_DBUS_LIB
+	auto lxqtStrategy = sm->sessionBus(Utils::lxqt, "org.lxqt.session", "/LXQtSession", "org.lxqt.session", "logout");
+
+	if (lxqtStrategy->notAvailable("canLogout"))
+		setEnabled(false, "No LXQt session found");
+	#endif // QT_DBUS_LIB
+
+	// info
+
+	sm->dummy(Utils::windows, "WinAPI ::ExitWindowsEx");
+
+/* TODO:
+	if (Utils::kde) {
+		setWarningStatus(i18n("You will be asked to save your documents."));
+	}
+*/
 }
 
 // RebootAction
@@ -1233,31 +834,20 @@ LogoutAction::LogoutAction() :
 // public
 
 RebootAction::RebootAction() :
-	StandardAction(
-		getDisplayName(ActionType::REBOOT),
-		QString::null, "reboot", U_SHUTDOWN_TYPE_REBOOT),
-	m_bootEntryComboBox(nullptr) {
-
-	if (!Config::oldActionNamesVar)
-		uiAction()->setStatusTip(i18n("Restart Computer"));
+	ExitAction(
+		i18n("Restart"),
+		QString(), "reboot", U_SHUTDOWN_TYPE_REBOOT
+	)
+	#ifdef Q_OS_LINUX
+	,
+// TODO: better help or example
+	systemOption("system", i18n("Select an Operating System to boot (used with --reboot)"), "entry")
+	#endif // Q_OS_LINUX
+{
+	uiAction()->setToolTip(i18n("Restart Computer"));
 
+// TODO: Utils::fromTheme(QString &name, QString &fallbackName) #api
 #ifdef KS_KF5
-/* HACK: crash on KDE 4.5.0
-	QPixmap p = KIconLoader::global()->loadIcon(
-		"system-reboot",
-		//"dialog-ok",
-		KIconLoader::NoGroup,
-		0, // default size
-		KIconLoader::DefaultState,
-		QStringList(), // no emblems
-		0L, // no path store
-		true // return "null" if no icon
-	);
-	if (p.isNull())
-		setIcon(QIcon::fromTheme("system-restart"));
-	else
-		setIcon(p);
-*/
 	// NOTE: follow the Icon Naming Specification and use "system-reboot" instead of "system-restart"
 	// <http://standards.freedesktop.org/icon-naming-spec/icon-naming-spec-latest.html>
 	uiAction()->setIcon(QIcon::fromTheme("system-reboot"));
@@ -1270,20 +860,70 @@ RebootAction::RebootAction() :
 #endif // KS_KF5
 
 	setCommandLineOption({ "r", "reboot", "restart" });
-	
+	#ifdef Q_OS_LINUX
+	CLI::getArgs()->addOption(systemOption);
+	#endif // Q_OS_LINUX
+
 	#ifdef QT_DBUS_LIB
+/* FIXME: blocking
+	auto sm = strategyManager();
+
+	if (m_kdeShutdownInterface != nullptr)
+		sm->sessionBus(Utils::kde, m_kdeShutdownInterface, "logoutAndReboot");
+*/
 	checkAvailable("CanRestart");
 	#endif // QT_DBUS_LIB
+
+	m_bootEntryComboBox = new BootEntryComboBox();
 }
 
-QWidget *RebootAction::getWidget() {
-/* TODO: boot entry combo box
-	#ifdef Q_OS_LINUX
-	if (!m_bootEntryComboBox)
-		m_bootEntryComboBox = new BootEntryComboBox();
-	#endif // Q_OS_LINUX
+void RebootAction::initContainerWidget() {
+	ExitAction::initContainerWidget();
+
+	auto *layout = getContainerForm();
+	layout->addRow(i18n("System:"), m_bootEntryComboBox);
+
+	updateBootEntryView();
+}
+
+bool RebootAction::onAction() {
+	if (m_bootEntryComboBox->currentIndex() > 0) {
+		BootEntry::setDefault(m_bootEntryComboBox->currentText());
+	}
+
+/* TEST:
+	qDebug() << "DRY RUN";
+// TODO: --dry-run CLI option
+	return true;
 */
-	return m_bootEntryComboBox;
+	return ExitAction::onAction();
+}
+
+bool RebootAction::onCommandLineOption() {
+	#ifdef Q_OS_LINUX
+	QString entry = CLI::getArgs()->value(systemOption);
+
+	if (entry.isEmpty()) {
+		m_bootEntryComboBox->setCurrentIndex(0); // default
+	}
+	else {
+// TODO: error message or fallback if no match
+		m_bootEntryComboBox->setCurrentText(entry);
+	}
+	#endif // Q_OS_LINUX
+
+	return true;
+}
+
+void RebootAction::updateBootEntryView() {
+	m_bootEntryComboBox->updateBootEntryList();
+
+	bool visible = Config::bootEntriesVisibleInWindow.getBool();
+
+	getContainerForm()->labelForField(m_bootEntryComboBox)
+		->setVisible(visible);
+
+	m_bootEntryComboBox->setVisible(visible);
 }
 
 // ShutDownAction
@@ -1291,12 +931,11 @@ QWidget *RebootAction::getWidget() {
 // public
 
 ShutDownAction::ShutDownAction() :
-	StandardAction(
-		getDisplayName(ActionType::SHUT_DOWN),
+	ExitAction(
+		i18n("Shut Down"),
 		"system-shutdown", "shutdown", U_SHUTDOWN_TYPE_HALT
 ) {
-	if (!Config::oldActionNamesVar)
-		uiAction()->setStatusTip(i18n("Turn Off Computer"));
+	uiAction()->setToolTip(i18n("Turn Off Computer"));
 
 	setCommandLineOption({
 		"h", "halt",
@@ -1304,6 +943,12 @@ ShutDownAction::ShutDownAction() :
 	});
 
 	#ifdef QT_DBUS_LIB
+/* FIXME: blocking
+	auto sm = strategyManager();
+
+	if (m_kdeShutdownInterface != nullptr)
+		sm->sessionBus(Utils::kde, m_kdeShutdownInterface, "logoutAndShutdown");
+*/
 	checkAvailable("CanStop");
 	#endif // QT_DBUS_LIB
 }
