@@ -8,6 +8,6 @@ echo "Testing ${BIN} -> ${LOG}"
 valgrind \
 	--leak-check=full \
 	--log-file="${LOG}" \
-	--show-reachable=yes \
+	--show-leak-kinds=definite,possible \
 	--tool=memcheck \
 	"${BIN}" -style fusion
